@@ -1,26 +1,21 @@
 """CherryPy-based server for accessing NILM database via HTTP"""
 
-# Need absolute_import so that "import nilmdb" won't pull in
-# nilmdb.py, but will pull the nilmdb module instead.
-from __future__ import absolute_import
+import os
+import json
+import socket
+import traceback
+
+import psutil
+import cherrypy
+
 import nilmdb.server
-from nilmdb.utils.printf import *
+from nilmdb.utils.printf import sprintf
 from nilmdb.server.errors import NilmDBError
 from nilmdb.utils.time import string_to_timestamp
 
-import cherrypy
-import sys
-import os
-import socket
-import simplejson as json
-import decorator
-import psutil
-import traceback
-
 from nilmdb.server.serverutil import (
     chunked_response,
     response_type,
-    workaround_cp_bug_1200,
     exception_to_httperror,
     CORS_allow,
     json_to_request_params,
@@ -33,17 +28,15 @@ from nilmdb.server.serverutil import (
 # Add CORS_allow tool
 cherrypy.tools.CORS_allow = cherrypy.Tool('on_start_resource', CORS_allow)
 
-class NilmApp(object):
+
+class NilmApp():
     def __init__(self, db):
         self.db = db
 
+
 # CherryPy apps
 class Root(NilmApp):
     """Root application for NILM database"""
-
-    def __init__(self, db):
-        super(Root, self).__init__(db)
-
     # /
     @cherrypy.expose
     def index(self):
@@ -72,11 +65,14 @@ class Root(NilmApp):
         path = self.db.get_basepath()
         usage = psutil.disk_usage(path)
         dbsize = nilmdb.utils.du(path)
-        return { "path": path,
-                 "size": dbsize,
-                 "other": max(usage.used - dbsize, 0),
-                 "reserved": max(usage.total - usage.used - usage.free, 0),
-                 "free": usage.free }
+        return {
+            "path": path,
+            "size": dbsize,
+            "other": max(usage.used - dbsize, 0),
+            "reserved": max(usage.total - usage.used - usage.free, 0),
+            "free": usage.free
+        }
+
 
 class Stream(NilmApp):
     """Stream-specific operations"""
@@ -89,7 +85,8 @@ class Stream(NilmApp):
                 start = string_to_timestamp(start_param)
         except Exception:
             raise cherrypy.HTTPError("400 Bad Request", sprintf(
-                "invalid start (%s): must be a numeric timestamp", start_param))
+                "invalid start (%s): must be a numeric timestamp",
+                start_param))
         try:
             if end_param is not None:
                 end = string_to_timestamp(end_param)
@@ -109,7 +106,7 @@ class Stream(NilmApp):
     # /stream/list?path=/newton/prep&extended=1
     @cherrypy.expose
     @cherrypy.tools.json_out()
-    def list(self, path = None, layout = None, extended = None):
+    def list(self, path=None, layout=None, extended=None):
         """List all streams in the database.  With optional path or
         layout parameter, just list streams that match the given path
         or layout.
@@ -128,7 +125,7 @@ class Stream(NilmApp):
     @cherrypy.tools.json_in()
     @cherrypy.tools.json_out()
     @exception_to_httperror(NilmDBError, ValueError)
-    @cherrypy.tools.CORS_allow(methods = ["POST"])
+    @cherrypy.tools.CORS_allow(methods=["POST"])
     def create(self, path, layout):
         """Create a new stream in the database.  Provide path
         and one of the nilmdb.layout.layouts keys.
@@ -140,7 +137,7 @@ class Stream(NilmApp):
     @cherrypy.tools.json_in()
     @cherrypy.tools.json_out()
     @exception_to_httperror(NilmDBError)
-    @cherrypy.tools.CORS_allow(methods = ["POST"])
+    @cherrypy.tools.CORS_allow(methods=["POST"])
     def destroy(self, path):
         """Delete a stream.  Fails if any data is still present."""
         return self.db.stream_destroy(path)
@@ -150,7 +147,7 @@ class Stream(NilmApp):
     @cherrypy.tools.json_in()
     @cherrypy.tools.json_out()
     @exception_to_httperror(NilmDBError, ValueError)
-    @cherrypy.tools.CORS_allow(methods = ["POST"])
+    @cherrypy.tools.CORS_allow(methods=["POST"])
     def rename(self, oldpath, newpath):
         """Rename a stream."""
         return self.db.stream_rename(oldpath, newpath)
@@ -166,16 +163,16 @@ class Stream(NilmApp):
         try:
             data = self.db.stream_get_metadata(path)
         except nilmdb.server.nilmdb.StreamError as e:
-            raise cherrypy.HTTPError("404 Not Found", e.message)
+            raise cherrypy.HTTPError("404 Not Found", str(e))
         if key is None:  # If no keys specified, return them all
-            key = data.keys()
+            key = list(data.keys())
         elif not isinstance(key, list):
-            key = [ key ]
+            key = [key]
         result = {}
         for k in key:
             if k in data:
                 result[k] = data[k]
-            else: # Return "None" for keys with no matching value
+            else:  # Return "None" for keys with no matching value
                 result[k] = None
         return result
 
@@ -185,11 +182,9 @@ class Stream(NilmApp):
             try:
                 data = dict(json.loads(data))
             except TypeError as e:
-                raise NilmDBError("can't parse 'data' parameter: " + e.message)
+                raise NilmDBError("can't parse 'data' parameter: " + str(e))
         for key in data:
-            if not (isinstance(data[key], basestring) or
-                    isinstance(data[key], float) or
-                    isinstance(data[key], int)):
+            if not isinstance(data[key], (str, float, int)):
                 raise NilmDBError("metadata values must be a string or number")
         function(path, data)
 
@@ -198,7 +193,7 @@ class Stream(NilmApp):
     @cherrypy.tools.json_in()
     @cherrypy.tools.json_out()
     @exception_to_httperror(NilmDBError, LookupError)
-    @cherrypy.tools.CORS_allow(methods = ["POST"])
+    @cherrypy.tools.CORS_allow(methods=["POST"])
     def set_metadata(self, path, data):
         """Set metadata for the named stream, replacing any existing
         metadata.  Data can be json-encoded or a plain dictionary."""
@@ -209,7 +204,7 @@ class Stream(NilmApp):
     @cherrypy.tools.json_in()
     @cherrypy.tools.json_out()
     @exception_to_httperror(NilmDBError, LookupError, ValueError)
-    @cherrypy.tools.CORS_allow(methods = ["POST"])
+    @cherrypy.tools.CORS_allow(methods=["POST"])
     def update_metadata(self, path, data):
         """Set metadata for the named stream, replacing any existing
         metadata.  Data can be json-encoded or a plain dictionary."""
@@ -219,8 +214,8 @@ class Stream(NilmApp):
     @cherrypy.expose
     @cherrypy.tools.json_out()
     @exception_to_httperror(NilmDBError, ValueError)
-    @cherrypy.tools.CORS_allow(methods = ["PUT"])
-    def insert(self, path, start, end, binary = False):
+    @cherrypy.tools.CORS_allow(methods=["PUT"])
+    def insert(self, path, start, end, binary=False):
         """
         Insert new data into the database.  Provide textual data
         (matching the path's layout) as a HTTP PUT.
@@ -246,8 +241,11 @@ class Stream(NilmApp):
                                          "application/octet-stream for "
                                          "binary data, not " + content_type)
 
+        # Note that non-binary data is *not* decoded from bytes to string,
+        # but rather passed directly to stream_insert.
+
         # Check path and get layout
-        if len(self.db.stream_list(path = path)) != 1:
+        if len(self.db.stream_list(path=path)) != 1:
             raise cherrypy.HTTPError("404", "No such stream: " + path)
 
         # Check limits
@@ -264,10 +262,10 @@ class Stream(NilmApp):
     # /stream/remove?path=/newton/prep&start=1234567890.0&end=1234567899.0
     @cherrypy.expose
     @cherrypy.tools.json_in()
-    @cherrypy.tools.CORS_allow(methods = ["POST"])
+    @cherrypy.tools.CORS_allow(methods=["POST"])
     @chunked_response
     @response_type("application/x-json-stream")
-    def remove(self, path, start = None, end = None):
+    def remove(self, path, start=None, end=None):
         """
         Remove data from the backend database.  Removes all data in
         the interval [start, end).
@@ -279,15 +277,15 @@ class Stream(NilmApp):
         """
         (start, end) = self._get_times(start, end)
 
-        if len(self.db.stream_list(path = path)) != 1:
+        if len(self.db.stream_list(path=path)) != 1:
             raise cherrypy.HTTPError("404", "No such stream: " + path)
 
-        @workaround_cp_bug_1200
         def content(start, end):
             # Note: disable chunked responses to see tracebacks from here.
             while True:
                 (removed, restart) = self.db.stream_remove(path, start, end)
-                yield json.dumps(removed) + "\r\n"
+                response = json.dumps(removed) + "\r\n"
+                yield response.encode('utf-8')
                 if restart is None:
                     break
                 start = restart
@@ -299,7 +297,7 @@ class Stream(NilmApp):
     @cherrypy.expose
     @chunked_response
     @response_type("application/x-json-stream")
-    def intervals(self, path, start = None, end = None, diffpath = None):
+    def intervals(self, path, start=None, end=None, diffpath=None):
         """
         Get intervals from backend database.  Streams the resulting
         intervals as JSON strings separated by CR LF pairs.  This may
@@ -316,20 +314,19 @@ class Stream(NilmApp):
         """
         (start, end) = self._get_times(start, end)
 
-        if len(self.db.stream_list(path = path)) != 1:
+        if len(self.db.stream_list(path=path)) != 1:
             raise cherrypy.HTTPError("404", "No such stream: " + path)
 
-        if diffpath and len(self.db.stream_list(path = diffpath)) != 1:
+        if diffpath and len(self.db.stream_list(path=diffpath)) != 1:
             raise cherrypy.HTTPError("404", "No such stream: " + diffpath)
 
-        @workaround_cp_bug_1200
         def content(start, end):
             # Note: disable chunked responses to see tracebacks from here.
             while True:
                 (ints, restart) = self.db.stream_intervals(path, start, end,
                                                            diffpath)
-                response = ''.join([ json.dumps(i) + "\r\n" for i in ints ])
-                yield response
+                response = ''.join([json.dumps(i) + "\r\n" for i in ints])
+                yield response.encode('utf-8')
                 if restart is None:
                     break
                 start = restart
@@ -338,8 +335,8 @@ class Stream(NilmApp):
     # /stream/extract?path=/newton/prep&start=1234567890.0&end=1234567899.0
     @cherrypy.expose
     @chunked_response
-    def extract(self, path, start = None, end = None,
-                count = False, markup = False, binary = False):
+    def extract(self, path, start=None, end=None,
+                count=False, markup=False, binary=False):
         """
         Extract data from backend database.  Streams the resulting
         entries as ASCII text lines separated by newlines.  This may
@@ -363,7 +360,7 @@ class Stream(NilmApp):
         (start, end) = self._get_times(start, end)
 
         # Check path and get layout
-        if len(self.db.stream_list(path = path)) != 1:
+        if len(self.db.stream_list(path=path)) != 1:
             raise cherrypy.HTTPError("404", "No such stream: " + path)
 
         if binary:
@@ -375,19 +372,18 @@ class Stream(NilmApp):
             content_type = "text/plain"
         cherrypy.response.headers['Content-Type'] = content_type
 
-        @workaround_cp_bug_1200
         def content(start, end):
             # Note: disable chunked responses to see tracebacks from here.
             if count:
                 matched = self.db.stream_extract(path, start, end,
-                                                 count = True)
-                yield sprintf("%d\n", matched)
+                                                 count=True)
+                yield sprintf(b"%d\n", matched)
                 return
 
             while True:
                 (data, restart) = self.db.stream_extract(
-                    path, start, end, count = False,
-                    markup = markup, binary = binary)
+                    path, start, end, count=False,
+                    markup=markup, binary=binary)
                 yield data
 
                 if restart is None:
@@ -395,29 +391,31 @@ class Stream(NilmApp):
                 start = restart
         return content(start, end)
 
-class Exiter(object):
+
+class Exiter():
     """App that exits the server, for testing"""
     @cherrypy.expose
     def index(self):
         cherrypy.response.headers['Content-Type'] = 'text/plain'
-        def content():
-            yield 'Exiting by request'
-            raise SystemExit
-        return content()
-    index._cp_config = { 'response.stream': True }
 
-class Server(object):
-    def __init__(self, db, host = '127.0.0.1', port = 8080,
-                 stoppable = False,       # whether /exit URL exists
-                 embedded = True,         # hide diagnostics and output, etc
-                 fast_shutdown = False,   # don't wait for clients to disconn.
-                 force_traceback = False, # include traceback in all errors
-                 basepath = '',           # base URL path for cherrypy.tree
+        def content():
+            yield b'Exiting by request'
+            raise SystemExit
+
+        return content()
+    index._cp_config = {'response.stream': True}
+
+
+class Server():
+    def __init__(self, db, host='127.0.0.1', port=8080,
+                 stoppable=False,        # whether /exit URL exists
+                 fast_shutdown=False,    # don't wait for clients to disconn.
+                 force_traceback=False,  # include traceback in all errors
+                 basepath='',            # base URL path for cherrypy.tree
                  ):
         # Save server version, just for verification during tests
         self.version = nilmdb.__version__
 
-        self.embedded = embedded
         self.db = db
         if not getattr(db, "_thread_safe", None):
             raise KeyError("Database object " + str(db) + " doesn't claim "
@@ -427,13 +425,12 @@ class Server(object):
 
         # Build up global server configuration
         cherrypy.config.update({
+            'environment': 'embedded',
             'server.socket_host': host,
             'server.socket_port': port,
             'engine.autoreload.on': False,
             'server.max_request_body_size': 8*1024*1024,
             })
-        if self.embedded:
-            cherrypy.config.update({ 'environment': 'embedded' })
 
         # Build up application specific configuration
         app_config = {}
@@ -442,23 +439,23 @@ class Server(object):
             })
 
         # Some default headers to just help identify that things are working
-        app_config.update({ 'response.headers.X-Jim-Is-Awesome': 'yeah' })
+        app_config.update({'response.headers.X-Jim-Is-Awesome': 'yeah'})
 
         # Set up Cross-Origin Resource Sharing (CORS) handler so we
         # can correctly respond to browsers' CORS preflight requests.
         # This also limits verbs to GET and HEAD by default.
-        app_config.update({ 'tools.CORS_allow.on': True,
-                            'tools.CORS_allow.methods': ['GET', 'HEAD'] })
+        app_config.update({'tools.CORS_allow.on': True,
+                           'tools.CORS_allow.methods': ['GET', 'HEAD']})
 
         # Configure the 'json_in' tool to also allow other content-types
         # (like x-www-form-urlencoded), and to treat JSON as a dict that
         # fills requests.param.
-        app_config.update({ 'tools.json_in.force': False,
-                            'tools.json_in.processor': json_to_request_params })
+        app_config.update({'tools.json_in.force': False,
+                           'tools.json_in.processor': json_to_request_params})
 
         # Send tracebacks in error responses.  They're hidden by the
         # error_page function for client errors (code 400-499).
-        app_config.update({ 'request.show_tracebacks' : True })
+        app_config.update({'request.show_tracebacks': True})
         self.force_traceback = force_traceback
 
         # Patch CherryPy error handler to never pad out error messages.
@@ -472,13 +469,12 @@ class Server(object):
         if stoppable:
             root.exit = Exiter()
         cherrypy.tree.apps = {}
-        cherrypy.tree.mount(root, basepath, config = { "/" : app_config })
+        cherrypy.tree.mount(root, basepath, config={"/": app_config})
 
         # Shutdowns normally wait for clients to disconnect.  To speed
         # up tests, set fast_shutdown = True
         if fast_shutdown:
-            # Setting timeout to 0 triggers os._exit(70) at shutdown, grr...
-            cherrypy.server.shutdown_timeout = 0.01
+            cherrypy.server.shutdown_timeout = 0
         else:
             cherrypy.server.shutdown_timeout = 5
 
@@ -490,18 +486,21 @@ class Server(object):
         return json_error_page(status, message, traceback, version,
                                self.force_traceback)
 
-    def start(self, blocking = False, event = None):
-        cherrypy_start(blocking, event, self.embedded)
+    def start(self, blocking=False, event=None):
+        cherrypy_start(blocking, event)
 
     def stop(self):
         cherrypy_stop()
 
+
 # Use a single global nilmdb.server.NilmDB and nilmdb.server.Server
 # instance since the database can only be opened once.  For this to
 # work, the web server must use only a single process and single
 # Python interpreter.  Multiple threads are OK.
 _wsgi_server = None
-def wsgi_application(dbpath, basepath): # pragma: no cover
+
+
+def wsgi_application(dbpath, basepath):
     """Return a WSGI application object with a database at the
     specified path.
 
@@ -516,33 +515,31 @@ def wsgi_application(dbpath, basepath): # pragma: no cover
         if _wsgi_server is None:
             # Try to start the server
             try:
-                db = nilmdb.utils.serializer_proxy(nilmdb.server.NilmDB)(dbpath)
+                db = nilmdb.utils.serializer_proxy(
+                    nilmdb.server.NilmDB)(dbpath)
                 _wsgi_server = nilmdb.server.Server(
-                    db, embedded = True,
-                    basepath = basepath.rstrip('/'))
+                    db, basepath=basepath.rstrip('/'))
             except Exception:
                 # Build an error message on failure
                 import pprint
                 err = sprintf("Initializing database at path '%s' failed:\n\n",
                               dbpath)
                 err += traceback.format_exc()
-                try:
-                    import pwd
-                    import grp
-                    err += sprintf("\nRunning as: uid=%d (%s), gid=%d (%s) "
-                                   "on host %s, pid %d\n",
-                                   os.getuid(), pwd.getpwuid(os.getuid())[0],
-                                   os.getgid(), grp.getgrgid(os.getgid())[0],
-                                   socket.gethostname(), os.getpid())
-                except ImportError:
-                    pass
+                import pwd
+                import grp
+                err += sprintf("\nRunning as: uid=%d (%s), gid=%d (%s) "
+                               "on host %s, pid %d\n",
+                               os.getuid(), pwd.getpwuid(os.getuid())[0],
+                               os.getgid(), grp.getgrgid(os.getgid())[0],
+                               socket.gethostname(), os.getpid())
                 err += sprintf("\nEnvironment:\n%s\n", pprint.pformat(environ))
         if _wsgi_server is None:
             # Serve up the error with our own mini WSGI app.
-            headers = [ ('Content-type', 'text/plain'),
-                        ('Content-length', str(len(err))) ]
+            err_b = err.encode('utf-8')
+            headers = [('Content-type', 'text/plain; charset=utf-8'),
+                       ('Content-length', str(len(err_b)))]
             start_response("500 Internal Server Error", headers)
-            return [err]
+            return [err_b]
 
         # Call the normal application
         return _wsgi_server.wsgi_application(environ, start_response)
