@@ -2,6 +2,8 @@
 
 # Iterator merging, based on http://code.activestate.com/recipes/491285/
 import heapq
+
+
 def imerge(*iterables):
     '''Merge multiple sorted inputs into a single sorted output.
 
@@ -17,8 +19,8 @@ def imerge(*iterables):
     h_append = h.append
     for it in map(iter, iterables):
         try:
-            next = it.next
-            h_append([next(), next])
+            nexter = it.__next__
+            h_append([nexter(), nexter])
         except _Stop:
             pass
     heapq.heapify(h)
@@ -26,9 +28,9 @@ def imerge(*iterables):
     while 1:
         try:
             while 1:
-                v, next = s = h[0]      # raises IndexError when h is empty
+                v, nexter = s = h[0]    # raises IndexError when h is empty
                 yield v
-                s[0] = next()           # raises StopIteration when exhausted
+                s[0] = nexter()         # raises StopIteration when exhausted
                 siftup(h, 0)            # restore heap condition
         except _Stop:
             heappop(h)                  # remove empty iterator
