@@ -5,8 +5,9 @@ import nilmdb.client
 
 from nilmdb.utils.printf import *
 from nilmdb.utils import timestamper
-from nilmdb.client import ClientError, ServerError
-from nilmdb.utils import datetime_tz
+from nilmdb.client import ClientError, ServerError, Error
+from nilmdb.utils.sort import sort_human
+import datetime_tz
 
 from nose.plugins.skip import SkipTest
 from nose.tools import *
@@ -16,8 +17,8 @@ import distutils.version
 import os
 import sys
 import threading
-import cStringIO
-import simplejson as json
+import io
+import json
 import unittest
 import warnings
 import resource
@@ -78,6 +79,32 @@ class TestClient(object):
         # Bad URLs should give 404, not 500
         with assert_raises(ClientError):
             client.http.get("/stream/create")
+
+        # Test error handling
+        url = testurl
+        args = { "url": url,
+                 "status": "400",
+                 "message": "Something went wrong",
+                 "traceback": None }
+        with assert_raises(ClientError):
+            client.http._handle_error(url, 400, json.dumps(args))
+        with assert_raises(ClientError):
+            client.http._handle_error(url, 400, "this is not JSON.. {")
+        args["status"] = "500"
+        with assert_raises(ServerError):
+            client.http._handle_error(url, 500, json.dumps(args))
+        args["message"] = None
+        with assert_raises(ServerError):
+            client.http._handle_error(url, 500, json.dumps(args))
+        args["status"] = "600"
+        with assert_raises(Error):
+            client.http._handle_error(url, 600, json.dumps(args))
+
+        # Use get_gen for an endpoint that doesn't have newlines,
+        # for better test coverage.
+        for line in client.http.get_gen("/version"):
+            pass
+
         client.close()
 
     def test_client_02_createlist(self):
@@ -98,7 +125,7 @@ class TestClient(object):
 
         # Bad method types
         with assert_raises(ClientError):
-            client.http.put("/stream/list","")
+            client.http.put("/stream/list",b"")
         # Try a bunch of times to make sure the request body is getting consumed
         for x in range(10):
             with assert_raises(ClientError):
@@ -111,6 +138,9 @@ class TestClient(object):
         client.stream_create("/newton/zzz/rawnotch2", "uint16_9")
         client.stream_create("/newton/zzz/rawnotch11", "uint16_9")
 
+        # Test sort_human (used by stream_list)
+        eq_(sort_human(["/s/10", "/s/2"]), ["/s/2", "/s/10"])
+
         # Verify we got 4 streams in the right order
         eq_(client.stream_list(), [ ["/newton/prep", "float32_8"],
                                     ["/newton/raw", "uint16_6"],
@@ -130,8 +160,19 @@ class TestClient(object):
         # we create a table.
         limit = resource.getrlimit(resource.RLIMIT_FSIZE)
         resource.setrlimit(resource.RLIMIT_FSIZE, (1, limit[1]))
+
+        # normal
         with assert_raises(ServerError) as e:
             client.stream_create("/newton/hello", "uint16_6")
+
+        # same but with force_traceback == False, to improve coverage
+        global test_server
+        test_server.force_traceback = False
+        with assert_raises(ServerError) as e:
+            client.stream_create("/newton/world", "uint16_6")
+        test_server.force_traceback = True
+
+        # Reset resource limit
         resource.setrlimit(resource.RLIMIT_FSIZE, limit)
 
         client.close()
@@ -197,7 +238,7 @@ class TestClient(object):
         data = timestamper.TimestamperRate(testfile, start, 120)
         with assert_raises(ClientError) as e:
             result = client.stream_insert("/newton/no-such-path", data)
-        in_("404 Not Found", str(e.exception))
+        in_("404 Not Found", repr(e.exception))
 
         # Now try reversed timestamps
         data = timestamper.TimestamperRate(testfile, start, 120)
@@ -209,27 +250,27 @@ class TestClient(object):
              "start must precede end", str(e.exception))
 
         # Now try empty data (no server request made)
-        empty = cStringIO.StringIO("")
+        empty = io.StringIO("")
         data = timestamper.TimestamperRate(empty, start, 120)
         result = client.stream_insert("/newton/prep", data)
         eq_(result, None)
 
         # It's OK to insert an empty interval
-        client.http.put("stream/insert", "", { "path": "/newton/prep",
+        client.http.put("stream/insert", b"", { "path": "/newton/prep",
                                                "start": 1, "end": 2 })
         eq_(list(client.stream_intervals("/newton/prep")), [[1, 2]])
         client.stream_remove("/newton/prep")
         eq_(list(client.stream_intervals("/newton/prep")), [])
 
         # Timestamps can be negative too
-        client.http.put("stream/insert", "", { "path": "/newton/prep",
+        client.http.put("stream/insert", b"", { "path": "/newton/prep",
                                                "start": -2, "end": -1 })
         eq_(list(client.stream_intervals("/newton/prep")), [[-2, -1]])
         client.stream_remove("/newton/prep")
         eq_(list(client.stream_intervals("/newton/prep")), [])
 
         # Intervals that end at zero shouldn't be any different
-        client.http.put("stream/insert", "", { "path": "/newton/prep",
+        client.http.put("stream/insert", b"", { "path": "/newton/prep",
                                                "start": -1, "end": 0 })
         eq_(list(client.stream_intervals("/newton/prep")), [[-1, 0]])
         client.stream_remove("/newton/prep")
@@ -237,38 +278,37 @@ class TestClient(object):
 
         # Try forcing a server request with equal start and end
         with assert_raises(ClientError) as e:
-            client.http.put("stream/insert", "", { "path": "/newton/prep",
+            client.http.put("stream/insert", b"", { "path": "/newton/prep",
                                                    "start": 0, "end": 0 })
         in_("400 Bad Request", str(e.exception))
         in_("start must precede end", str(e.exception))
 
         # Invalid times in HTTP request
         with assert_raises(ClientError) as e:
-            client.http.put("stream/insert", "", { "path": "/newton/prep",
+            client.http.put("stream/insert", b"", { "path": "/newton/prep",
                                                    "start": "asdf", "end": 0 })
         in_("400 Bad Request", str(e.exception))
         in_("invalid start", str(e.exception))
 
         with assert_raises(ClientError) as e:
-            client.http.put("stream/insert", "", { "path": "/newton/prep",
+            client.http.put("stream/insert", b"", { "path": "/newton/prep",
                                                    "start": 0, "end": "asdf" })
         in_("400 Bad Request", str(e.exception))
         in_("invalid end", str(e.exception))
 
         # Good content type
         with assert_raises(ClientError) as e:
-            client.http.put("stream/insert", "",
+            client.http.put("stream/insert", b"",
                             { "path": "xxxx", "start": 0, "end": 1,
-                              "binary": 1 },
-                            binary = True)
+                              "binary": 1 })
         in_("No such stream", str(e.exception))
 
         # Bad content type
         with assert_raises(ClientError) as e:
-            client.http.put("stream/insert", "",
+            client.http.put("stream/insert", b"",
                             { "path": "xxxx", "start": 0, "end": 1,
                               "binary": 1 },
-                            binary = False)
+                            content_type="text/plain; charset=utf-8")
         in_("Content type must be application/octet-stream", str(e.exception))
 
         # Specify start/end (starts too late)
@@ -292,22 +332,32 @@ class TestClient(object):
                          ">= end time 1332511201000000", str(e.exception))
                is not None)
 
+        def check_data():
+            # Verify the intervals.  Should be just one, even if the data
+            # was inserted in chunks, due to nilmdb interval concatenation.
+            intervals = list(client.stream_intervals("/newton/prep"))
+            eq_(intervals, [[start, start + 119999777]])
+
+            # Try some overlapping data -- just insert it again
+            data = timestamper.TimestamperRate(testfile, start, 120)
+            with assert_raises(ClientError) as e:
+                result = client.stream_insert("/newton/prep", data)
+            in_("400 Bad Request", str(e.exception))
+            in_("verlap", str(e.exception))
+
         # Now do the real load
         data = timestamper.TimestamperRate(testfile, start, 120)
         result = client.stream_insert("/newton/prep", data,
                                       start, start + 119999777)
+        check_data()
 
-        # Verify the intervals.  Should be just one, even if the data
-        # was inserted in chunks, due to nilmdb interval concatenation.
-        intervals = list(client.stream_intervals("/newton/prep"))
-        eq_(intervals, [[start, start + 119999777]])
-
-        # Try some overlapping data -- just insert it again
+        # Try inserting directly-passed data
+        client.stream_remove("/newton/prep", start, start + 119999777)
         data = timestamper.TimestamperRate(testfile, start, 120)
-        with assert_raises(ClientError) as e:
-            result = client.stream_insert("/newton/prep", data)
-        in_("400 Bad Request", str(e.exception))
-        in_("verlap", str(e.exception))
+        data_bytes = b''.join(data)
+        result = client.stream_insert("/newton/prep", data_bytes,
+                                      start, start + 119999777)
+        check_data()
 
         nilmdb.client.client.StreamInserter._max_data = old_max_data
         client.close()
@@ -333,7 +383,7 @@ class TestClient(object):
         with assert_raises(ClientError) as e:
             list(client.stream_extract("/newton/prep",
                                        count = True, binary = True))
-        data = "".join(client.stream_extract("/newton/prep", binary = True))
+        data = b"".join(client.stream_extract("/newton/prep", binary = True))
         # Quick check using struct
         unpacker = struct.Struct("<qffffffff")
         out = []
@@ -359,18 +409,18 @@ class TestClient(object):
         end = nilmdb.utils.time.parse_time("20120323T1000")
         for function in [ client.stream_intervals, client.stream_extract ]:
             with assert_raises(ClientError) as e:
-                function("/newton/prep", start, end).next()
+                next(function("/newton/prep", start, end))
             in_("400 Bad Request", str(e.exception))
             in_("start must precede end", str(e.exception))
 
         # Trigger a curl error in generator
         with assert_raises(ServerError) as e:
-            client.http.get_gen("http://nosuchurl.example.com./").next()
+            next(client.http.get_gen("http://nosuchurl.example.com./"))
 
         # Check 404 for missing streams
         for function in [ client.stream_intervals, client.stream_extract ]:
             with assert_raises(ClientError) as e:
-                function("/no/such/stream").next()
+                next(function("/no/such/stream"))
             in_("404 Not Found", str(e.exception))
             in_("No such stream", str(e.exception))
 
@@ -391,7 +441,7 @@ class TestClient(object):
 
         def headers():
             h = ""
-            for (k, v) in http._last_response.headers.items():
+            for (k, v) in list(http._last_response.headers.items()):
                 h += k + ": " + v + "\n"
             return h.lower()
 
@@ -454,8 +504,8 @@ class TestClient(object):
             eq_(client.stream_list(), [])
 
             # Create Unicode stream, match it
-            raw = [ u"/düsseldorf/raw", u"uint16_6" ]
-            prep = [ u"/düsseldorf/prep", u"uint16_6" ]
+            raw = [ "/düsseldorf/raw", "uint16_6" ]
+            prep = [ "/düsseldorf/prep", "uint16_6" ]
             client.stream_create(*raw)
             eq_(client.stream_list(), [raw])
             eq_(client.stream_list(layout=raw[1]), [raw])
@@ -466,10 +516,10 @@ class TestClient(object):
             # Set / get metadata with Unicode keys and values
             eq_(client.stream_get_metadata(raw[0]), {})
             eq_(client.stream_get_metadata(prep[0]), {})
-            meta1 = { u"alpha": u"α",
-                      u"β": u"beta" }
-            meta2 = { u"alpha": u"α" }
-            meta3 = { u"β": u"beta" }
+            meta1 = { "alpha": "α",
+                      "β": "beta" }
+            meta2 = { "alpha": "α" }
+            meta3 = { "β": "beta" }
             client.stream_set_metadata(prep[0], meta1)
             client.stream_update_metadata(prep[0], {})
             client.stream_update_metadata(raw[0], meta2)
@@ -513,68 +563,74 @@ class TestClient(object):
             # override _max_data to trigger frequent server updates
             ctx._max_data = 15
 
-            ctx.insert("1000 1\n")
+            ctx.insert(b"1000 1\n")
 
-            ctx.insert("1010 ")
-            ctx.insert("1\n1020 1")
-            ctx.insert("")
-            ctx.insert("\n1030 1\n")
+            ctx.insert(b"1010 ")
+            ctx.insert(b"1\n1020 1")
+            ctx.insert(b"")
+            ctx.insert(b"\n1030 1\n")
 
-            ctx.insert("1040 1\n")
-            ctx.insert("# hello\n")
-            ctx.insert("   # hello\n")
-            ctx.insert("  1050 1\n")
+            ctx.insert(b"1040 1\n")
+            ctx.insert(b"# hello\n")
+            ctx.insert(b"   # hello\n")
+            ctx.insert(b"  1050 1\n")
             ctx.finalize()
 
-            ctx.insert("1070 1\n")
+            ctx.insert(b"1070 1\n")
             ctx.update_end(1080)
             ctx.finalize()
             ctx.update_start(1090)
-            ctx.insert("1100 1\n")
-            ctx.insert("1110 1\n")
+            ctx.insert(b"1100 1\n")
+            ctx.insert(b"1110 1\n")
             ctx.send()
-            ctx.insert("1120 1\n")
-            ctx.insert("1130 1\n")
-            ctx.insert("1140 1\n")
+            ctx.insert(b"1120 1\n")
+            ctx.insert(b"1130 1\n")
+            ctx.insert(b"1140 1\n")
             ctx.update_end(1160)
-            ctx.insert("1150 1\n")
+            ctx.insert(b"1150 1\n")
             ctx.update_end(1170)
-            ctx.insert("1160 1\n")
+            ctx.insert(b"1160 1\n")
             ctx.update_end(1180)
-            ctx.insert("1170 1" +
-                       " # this is super long" * 100 +
-                       "\n")
+            ctx.insert(b"1170 1" +
+                       b" # this is super long" * 100 +
+                       b"\n")
             ctx.finalize()
-            ctx.insert("# this is super long" * 100)
+            ctx.insert(b"# this is super long" * 100)
+
+            # override _max_data_after_send to trigger ValueError on a
+            # long nonterminated line
+            ctx._max_data_after_send = 1000
+            with assert_raises(ValueError):
+                ctx.insert(b"# this is super long" * 100)
 
         with assert_raises(ClientError):
             with client.stream_insert_context("/context/test",
                                               1000, 2000) as ctx:
-                ctx.insert("1180 1\n")
+                ctx.insert(b"1180 1\n")
 
         with assert_raises(ClientError):
             with client.stream_insert_context("/context/test",
                                               2000, 3000) as ctx:
-                ctx.insert("1180 1\n")
+                ctx.insert(b"1180 1\n")
 
         with assert_raises(ClientError):
             with client.stream_insert_context("/context/test") as ctx:
-                ctx.insert("bogus data\n")
+                ctx.insert(b"bogus data\n")
 
         with client.stream_insert_context("/context/test", 2000, 3000) as ctx:
             # make sure our override wasn't permanent
             ne_(ctx._max_data, 15)
-            ctx.insert("2250 1\n")
+            ctx.insert(b"2250 1\n")
             ctx.finalize()
 
         with assert_raises(ClientError):
             with client.stream_insert_context("/context/test",
                                               3000, 4000) as ctx:
-                ctx.insert("3010 1\n")
-                ctx.insert("3020 2\n")
-                ctx.insert("3030 3\n")
-                ctx.insert("3040 4\n")
-                ctx.insert("3040 4\n") # non-monotonic after a few lines
+                ctx.insert(b"3010 1\n")
+                ctx.insert(b"3020 2\n")
+                ctx.insert(b"3030 3\n")
+                ctx.insert(b"3040 4\n")
+                ctx.insert(b"3040 4\n") # non-monotonic after a few lines
                 ctx.finalize()
 
         eq_(list(client.stream_intervals("/context/test")),
@@ -609,9 +665,9 @@ class TestClient(object):
         # Insert a region with just a few points
         with client.stream_insert_context("/empty/test") as ctx:
             ctx.update_start(100)
-            ctx.insert("140 1\n")
-            ctx.insert("150 1\n")
-            ctx.insert("160 1\n")
+            ctx.insert(b"140 1\n")
+            ctx.insert(b"150 1\n")
+            ctx.insert(b"160 1\n")
             ctx.update_end(200)
             ctx.finalize()
 
@@ -624,7 +680,7 @@ class TestClient(object):
 
         # Try also creating a completely empty interval from scratch,
         # in a few different ways.
-        client.stream_insert("/empty/test", "", 300, 350)
+        client.stream_insert("/empty/test", b"", 300, 350)
         client.stream_insert("/empty/test", [], 400, 450)
         with client.stream_insert_context("/empty/test", 500, 550):
             pass
@@ -649,10 +705,10 @@ class TestClient(object):
             ctx.finalize() # inserts [1000, 1050]
             ctx.finalize() # nothing
             ctx.finalize() # nothing
-            ctx.insert("1100 1\n")
+            ctx.insert(b"1100 1\n")
             ctx.finalize() # inserts [1100, 1101]
             ctx.update_start(1199)
-            ctx.insert("1200 1\n")
+            ctx.insert(b"1200 1\n")
             ctx.update_end(1250)
             ctx.finalize() # inserts [1199, 1250]
             ctx.update_start(1299)
@@ -660,14 +716,14 @@ class TestClient(object):
             ctx.update_end(1350)
             ctx.finalize() # nothing
             ctx.update_start(1400)
-            ctx.insert("# nothing!\n")
+            ctx.insert(b"# nothing!\n")
             ctx.update_end(1450)
             ctx.finalize()
             ctx.update_start(1500)
-            ctx.insert("# nothing!")
+            ctx.insert(b"# nothing!")
             ctx.update_end(1550)
             ctx.finalize()
-            ctx.insert("# nothing!\n" * 10)
+            ctx.insert(b"# nothing!\n" * 10)
             ctx.finalize()
             # implicit last finalize inserts [1400, 1450]
 
@@ -709,19 +765,19 @@ class TestClient(object):
         client.stream_create("/rounding/test", "uint16_1")
         with client.stream_insert_context("/rounding/test",
                                           100000000, 200000000.1) as ctx:
-            ctx.insert("100000000.1 1\n")
-            ctx.insert("150000000.00003 1\n")
-            ctx.insert("199999999.4 1\n")
+            ctx.insert(b"100000000.1 1\n")
+            ctx.insert(b"150000000.00003 1\n")
+            ctx.insert(b"199999999.4 1\n")
         eq_(list(client.stream_intervals("/rounding/test")),
             [ [ 100000000, 200000000 ] ])
 
         with assert_raises(ClientError):
             with client.stream_insert_context("/rounding/test",
                                               200000000, 300000000) as ctx:
-                ctx.insert("200000000 1\n")
-                ctx.insert("250000000 1\n")
+                ctx.insert(b"200000000 1\n")
+                ctx.insert(b"250000000 1\n")
                 # Server will round this and give an error on finalize()
-                ctx.insert("299999999.99 1\n")
+                ctx.insert(b"299999999.99 1\n")
 
         client.stream_remove("/rounding/test")
         client.stream_destroy("/rounding/test")
