@@ -4,7 +4,7 @@ import nilmdb.server
 
 from nilmdb.utils.printf import *
 import nilmdb.cmdline
-from nilmdb.utils import datetime_tz
+import datetime_tz
 
 import unittest
 from nose.tools import *
@@ -13,7 +13,7 @@ import itertools
 import os
 import re
 import sys
-import StringIO
+import io
 import shlex
 import warnings
 
@@ -54,11 +54,6 @@ def setup_module():
 def teardown_module():
     server_stop()
 
-# Add an encoding property to StringIO so Python will convert Unicode
-# properly when writing or reading.
-class UTF8StringIO(StringIO.StringIO):
-    encoding = 'utf-8'
-
 class TestCmdline(object):
 
     def run(self, arg_string, infile=None, outfile=None):
@@ -76,23 +71,29 @@ class TestCmdline(object):
                 ( sys.stdin, sys.stdout, sys.stderr ) = self.saved
         # Empty input if none provided
         if infile is None:
-            infile = UTF8StringIO("")
+            infile = io.TextIOWrapper(io.BytesIO(b""))
         # Capture stderr
-        errfile = UTF8StringIO()
+        errfile = io.TextIOWrapper(io.BytesIO())
         if outfile is None:
             # If no output file, capture stdout with stderr
             outfile = errfile
         with stdio_wrapper(infile, outfile, errfile) as s:
             try:
-                # shlex doesn't support Unicode very well.  Encode the
-                # string as UTF-8 explicitly before splitting.
-                args = shlex.split(arg_string.encode('utf-8'))
+                args = shlex.split(arg_string)
                 nilmdb.cmdline.Cmdline(args).run()
                 sys.exit(0)
             except SystemExit as e:
                 exitcode = e.code
-        captured = nilmdb.utils.unicode.decode(outfile.getvalue())
-        self.captured = captured
+
+        # Capture raw binary output, and also try to decode a Unicode
+        # string copy.
+        self.captured_binary = outfile.buffer.getvalue()
+        try:
+            outfile.seek(0)
+            self.captured = outfile.read()
+        except UnicodeDecodeError:
+            self.captured = None
+
         self.exitcode = exitcode
 
     def ok(self, arg_string, infile = None):
@@ -131,16 +132,16 @@ class TestCmdline(object):
         with open(file) as f:
             contents = f.read()
             if contents != self.captured:
-                print "--- reference file (first 1000 bytes):\n"
-                print contents[0:1000] + "\n"
-                print "--- captured data (first 1000 bytes):\n"
-                print self.captured[0:1000] + "\n"
-                zipped = itertools.izip_longest(contents, self.captured)
+                print("--- reference file (first 1000 bytes):\n")
+                print(contents[0:1000] + "\n")
+                print("--- captured data (first 1000 bytes):\n")
+                print(self.captured[0:1000] + "\n")
+                zipped = itertools.zip_longest(contents, self.captured)
                 for (n, (a, b)) in enumerate(zipped):
                     if a != b:
-                        print "--- first difference is at offset", n
-                        print "--- reference:", repr(a)
-                        print "---  captured:", repr(b)
+                        print("--- first difference is at offset", n)
+                        print("--- reference:", repr(a))
+                        print("---  captured:", repr(b))
                         break
                 raise AssertionError("captured data doesn't match " + file)
 
@@ -359,6 +360,12 @@ class TestCmdline(object):
         self.ok("metadata /newton/raw --update "
                 "v_scale=1.234")
 
+        # unicode
+        self.ok("metadata /newton/raw --set "
+                "a_𝓴𝓮𝔂=value a_key=𝓿𝓪𝓵𝓾𝓮 a_𝗸𝗲𝘆=𝘃𝗮𝗹𝘂𝗲")
+        self.ok("metadata /newton/raw --get")
+        self.match("a_key=𝓿𝓪𝓵𝓾𝓮\na_𝓴𝓮𝔂=value\na_𝗸𝗲𝘆=𝘃𝗮𝗹𝘂𝗲\n")
+
         # various parsing tests
         self.ok("metadata /newton/raw --update foo=")
         self.fail("metadata /newton/raw --update =bar")
@@ -492,6 +499,10 @@ class TestCmdline(object):
         self.ok("insert -t -r 120 --start '03/23/2012 06:05:00' /newton/prep "
                 "/dev/null")
 
+        # --quiet option
+        self.ok("insert --quiet -t -r 120 -s @0 /newton/prep /dev/null")
+        self.match("")
+
         # bad start time
         self.fail("insert -t -r 120 --start 'whatever' /newton/prep /dev/null")
 
@@ -590,6 +601,13 @@ class TestCmdline(object):
                   exitcode = 2, require_error = False)
         self.contain("no data")
 
+        # unannotated empty extract is just empty, with an exit code of 2
+        self.fail("extract /newton/prep " +
+                  "--start '23 Mar 2022 10:00:30' " +
+                  "--end '23 Mar 2022 10:00:31'",
+                  exitcode = 2, require_error = False)
+        self.match("")
+
         # but are ok if we're just counting results
         self.ok("extract --count /newton/prep " +
                 "--start '23 Mar 2012 20:00:30' " +
@@ -600,6 +618,14 @@ class TestCmdline(object):
                 "--end '23 Mar 2012 20:00:30.000002'")
         self.match("0\n")
 
+        # Extract needs --start and --end
+        self.fail("extract -a /newton/prep")
+        self.contain("arguments are required")
+        self.fail("extract -a /newton/prep --start 2000-01-01")
+        self.contain("arguments are required")
+        self.fail("extract -a /newton/prep --end 2000-01-01")
+        self.contain("arguments are required")
+
         # Check various dumps against stored copies of how they should appear
         def test(file, start, end, extra=""):
             self.ok("extract " + extra + " /newton/prep " +
@@ -636,7 +662,7 @@ class TestCmdline(object):
         self.fail("extract -m -B /newton/prep -s min -e max")
         self.contain("binary cannot be combined")
         self.ok("extract -B /newton/prep -s min -e max")
-        eq_(len(self.captured), 43200 * (8 + 8*4))
+        eq_(len(self.captured_binary), 43200 * (8 + 8*4))
 
         # markup for 3 intervals, plus extra markup lines whenever we had
         # a "restart" from the nilmdb.stream_extract function
@@ -756,7 +782,7 @@ class TestCmdline(object):
         self.ok("destroy --help")
 
         self.fail("destroy")
-        self.contain("too few arguments")
+        self.contain("the following arguments are required")
 
         self.fail("destroy /no/such/stream")
         self.contain("no stream matched path")
@@ -818,18 +844,18 @@ class TestCmdline(object):
         self.ok("destroy /newton/prep /newton/raw")
         self.ok("destroy /newton/zzz")
 
-        self.ok(u"create /düsseldorf/raw uint16_6")
+        self.ok("create /düsseldorf/raw uint16_6")
         self.ok("list -l --detail")
-        self.contain(u"/düsseldorf/raw uint16_6")
+        self.contain("/düsseldorf/raw uint16_6")
         self.contain("(no intervals)")
 
         # Unicode metadata
-        self.ok(u"metadata /düsseldorf/raw --set α=beta 'γ=δ'")
-        self.ok(u"metadata /düsseldorf/raw --update 'α=β ε τ α'")
-        self.ok(u"metadata /düsseldorf/raw")
-        self.match(u"α=β ε τ α\nγ=δ\n")
+        self.ok("metadata /düsseldorf/raw --set α=beta 'γ=δ'")
+        self.ok("metadata /düsseldorf/raw --update 'α=β ε τ α'")
+        self.ok("metadata /düsseldorf/raw")
+        self.match("α=β ε τ α\nγ=δ\n")
 
-        self.ok(u"destroy /düsseldorf/raw")
+        self.ok("destroy /düsseldorf/raw")
 
     def test_13_files(self):
         # Test BulkData's ability to split into multiple files,
@@ -1146,3 +1172,74 @@ class TestCmdline(object):
 
         server_stop()
         server_start()
+
+    def test_05b_completion(self):
+        # Test bash completion.  This depends on some data put in the DB by
+        # earlier tests, so the execution order is important.
+        def complete(line, expect="<unspecified>"):
+            # set env vars
+            env = {
+                '_ARGCOMPLETE': '1',
+                'COMP_LINE': line,
+                'COMP_POINT': str(len(line)),
+                'COMP_TYPE': '8',
+                'NILMDB_URL': "http://localhost:32180/",
+            }
+            for (k, v) in env.items():
+                os.environ[k] = v
+
+            # create pipe for completion output
+            output = io.BytesIO()
+
+            # ensure argcomplete won't mess with any FDs
+            def fake_fdopen(fd, mode):
+                return io.BytesIO()
+            old_fdopen = os.fdopen
+            os.fdopen = fake_fdopen
+
+            # run cli
+            cmdline = nilmdb.cmdline.Cmdline([])
+            cmdline.complete_output_stream = output
+            try:
+                cmdline.run()
+                sys.exit(0)
+            except SystemExit as e:
+                exitcode = e.code
+            eq_(exitcode, 0)
+
+            # clean up
+            os.fdopen = old_fdopen
+            for (k, v) in env.items():
+                del os.environ[k]
+
+            # read completion output
+            comp = output.getvalue()
+
+            # replace completion separators with commas, for clarity
+            cleaned = comp.replace(b'\x0b', b',').decode('utf-8')
+
+            # expect the given match or prefix
+            if expect.endswith('*'):
+                if not cleaned.startswith(expect[:-1]):
+                    raise AssertionError(("completions:\n  '%s'\n"
+                                          "don't start with:\n  '%s'") %
+                                         (cleaned, expect[:-1]))
+            else:
+                if cleaned != expect:
+                    raise AssertionError(("completions:\n  '%s'\n"
+                                          "don't match:\n  '%s'") %
+                                         (cleaned, expect))
+
+        complete("nilmtool -u ", "")
+        complete("nilmtool list ", "-h,--help,-E,--ext*")
+        complete("nilmtool list --st", "--start ")
+        complete("nilmtool list --start ", "")
+        complete("nilmtool list /", "/newton/prep,/newton/raw*")
+        complete("nilmtool create /foo int3", "int32_1,int32_2*")
+        complete("nilmtool metadata /newton/raw --get a",
+                 "a_𝓴𝓮𝔂,a_key,a_𝗸𝗲𝘆")
+        complete("nilmtool metadata /newton/raw --set a",
+                 "a_𝓴𝓮𝔂=value,a_key=𝓿𝓪𝓵𝓾𝓮,a_𝗸𝗲𝘆=𝘃𝗮𝗹𝘂𝗲")
+        complete("nilmtool metadata /newton/raw --set a_𝗸", "a_𝗸𝗲𝘆=𝘃𝗮𝗹𝘂𝗲 ")
+        complete("nilmtool metadata '' --set a", "")
+        self.run("list")
